import PyInstaller.__main__
from pathlib import Path

from PyInstaller.utils.hooks import collect_submodules

HERE = Path(__file__).parent.absolute()
path_to_main = str(HERE / "__main__.py")

def install():
    PyInstaller.__main__.run([
        path_to_main,
        '--onefile',
        '--windowed',
        '--collect-all=vosk'
        # other pyinstaller options...
    ])



import asyncio
import datetime
import json
import logging
import re
import typing

import pyaudio
import numpy
from pydub.silence import detect_leading_silence
from pydub.utils import make_chunks

from py_speech_service import speech_service_pb2
from py_speech_service.piper import Piper

trim_leading_silence = lambda x: x[detect_leading_silence(x):]
trim_trailing_silence = lambda x: trim_leading_silence(x.reverse()).reverse()
strip_silence = lambda x: trim_trailing_silence(trim_leading_silence(x))

class SpeechSettings:
    model_name: str = "hfc_female"
    onnx_path: str = ""
    config_path: str = ""
    alt_model_name: str = "hfc_male"
    alt_onnx_path: str = ""
    alt_config_path: str = ""
    speed: float = 1
    pitch: float = 1
    gain: float = 0
    
    def __init__(self, other = None):
        if type(other) is SpeechSettings:
            self.model_name = other.model_name
            self.onnx_path = other.onnx_path
            self.config_path = other.config_path
            self.alt_model_name = other.alt_model_name
            self.alt_onnx_path = other.alt_onnx_path
            self.alt_config_path = other.alt_config_path
            self.speed = other.speed
            self.pitch = other.pitch
            self.gain = other.gain
        elif other is not None:
            self.model_name = other.model_name if hasattr(other, "model_name") and other.model_name else ""
            self.onnx_path = other.onnx_path if hasattr(other, "onnx_path") and other.onnx_path else ""
            self.config_path = other.config_path if hasattr(other, "config_path") and other.config_path else ""
            self.alt_model_name = other.alt_model_name if hasattr(other, "alt_model_name") and other.alt_model_name else ""
            self.alt_onnx_path = other.alt_onnx_path if hasattr(other, "alt_onnx_path") and other.alt_onnx_path else ""
            self.alt_config_path = other.alt_config_path if hasattr(other, "alt_config_path") and other.alt_config_path else ""
            self.speed = numpy.clip(other.speed, 0.5, 2.0) if hasattr(other, "speed") and other.speed else 1
            self.gain = numpy.clip(other.gain, -100.0, 100.0) if hasattr(other, "gain") and other.gain else 0
            self.pitch = numpy.clip(other.pitch, 0.5, 1.5) if hasattr(other, "pitch") and other.pitch else 1


class PendingSpeechRequest:
    message: str
    silence_seconds: float
    speech_settings: SpeechSettings
    use_alt_voice: bool

    original_message: str
    first_request_of_message: bool
    last_request_of_message: bool

    def to_string(self):
        data = {
            "message": str(self.message) if hasattr(self, "message") else "",
            "silence_seconds": str(self.silence_seconds) if hasattr(self, "silence_seconds") else "",
            "model_name": str(self.speech_settings.model_name) if hasattr(self, "model_name") else "",
            "onnx_path": str(self.speech_settings.onnx_path) if hasattr(self, "onnx_path") else "",
            "config_path": str(self.speech_settings.config_path) if hasattr(self, "config_path") else "",
            "use_alt_voice": str(self.use_alt_voice) if hasattr(self, "use_alt_voice") else "",
            "speed_modifier": str(self.speech_settings.speed) if hasattr(self, "speech_settings") else "",
            "gain_modifier": str(self.speech_settings.gain) if hasattr(self, "speech_settings") else "",
            "pitch_modifier": str(self.speech_settings.pitch) if hasattr(self, "speech_settings") else "",
        }
        return json.dumps(data)
    


class Speaker:

    speech_queue: asyncio.Queue[PendingSpeechRequest] = asyncio.Queue()
    grpc_response_queue: typing.Optional[asyncio.Queue] = None
    shutdown_event = asyncio.Event()
    stop_talking_event = asyncio.Event()

    default_settings: SpeechSettings = SpeechSettings()
    piper: Piper = Piper()

    def start(self):
        asyncio.create_task(self.process_speech_queue())

    async def speak_basic_line(self, line: str):
        asyncio.create_task(self.process_speech_queue())
        response_queue = asyncio.Queue()
        self.set_grpc_response_queue(response_queue)
        await self.speak(line)
        while not self.shutdown_event.is_set():  # Keep running unless shutdown is triggered
            try:
                response = await asyncio.wait_for(response_queue.get(), timeout=1.0)  # Wait for 1 second
                if hasattr(response.speak_update, "is_end_of_message") and response.speak_update.is_end_of_message:
                    break
            except asyncio.TimeoutError:
                continue  # Retry checking


    async def process_speech_queue(self):
        logging.info("Started processing speech queue")
        print("Started processing speech queue")
        while not self.shutdown_event.is_set():  # Keep running unless shutdown is triggered
            try:
                item = await asyncio.wait_for(self.speech_queue.get(), timeout=1.0)  # Wait for 1 second
                await self.__handle_request(item)
            except asyncio.TimeoutError:
                continue  # Retry checking
        logging.info("Stopped processing speech queue")
        print("Stopped processing speech queue")
        self.speech_queue.task_done()

    def set_grpc_response_queue(self, queue: asyncio.Queue):
        self.grpc_response_queue = queue

    async def speak(self, message: str, settings: typing.Optional[SpeechSettings] = None):

        request = PendingSpeechRequest()

        if settings is None:
            request.speech_settings = SpeechSettings(self.default_settings)
        else:
            request.speech_settings = SpeechSettings(settings)

        if not message.__contains__("</") and not message.__contains__("/>"):
            request.message = message
            request.original_message = message
            request.first_request_of_message = True
            request.last_request_of_message = True
            await self.speech_queue.put(request)
        else:
            parts = self.__split_by_tags(message)
            requests: list[PendingSpeechRequest] = []
            tags: list[str] = []

            for part in parts:
                if part == '':
                    continue
                if part.startswith("<break"):
                    requests.append(self.__create_ssml_request(request, part, tags))
                elif part.startswith("</"):
                    tags.pop()
                elif part.startswith("<"):
                    tags.append(part)
                else:
                    requests.append(self.__create_ssml_request(request, part, tags))

            is_first: bool = True
            last_request = requests[len(requests)-1]
            for request in requests:
                request.first_request_of_message = is_first
                request.last_request_of_message = request == last_request
                request.original_message = message
                is_first = False
                await self.speech_queue.put(request)

    def stop_speaking(self):
        try:
            while not self.speech_queue.empty():
                self.speech_queue.get_nowait()
                self.speech_queue.task_done()
        except Exception as e:
            logging.info("Cleared speech queue")
            print("Cleared speech queue")

        self.stop_talking_event.set()

    def shutdown(self):
        self.shutdown_event.set()

    @staticmethod
    def __split_by_tags(html: str) -> list[str]:
        pattern = r"(</?[^>]+>)|([^<]+)"
        matches = re.findall(pattern, html)
        result = [match[0].strip() or match[1].strip() for match in matches if match[0] or match[1]]
        return result

    @staticmethod
    def __extract_attributes(tag) -> dict[str, str]:
        pattern = r'(\w+)=["\']([^"\']*?)["\']'
        attributes = dict(re.findall(pattern, tag))
        return attributes

    def __create_ssml_request(self, initial_request: PendingSpeechRequest, message: str, prosody_tags: list[str]) -> PendingSpeechRequest:
        request = PendingSpeechRequest()
        request.speech_settings = SpeechSettings(initial_request.speech_settings)

        if message.startswith("<break"):
            attr = self.__extract_attributes(message)
            if attr.__contains__("time"):
                if attr["time"].endswith("ms"):
                    request.silence_seconds = float(attr["time"].replace("ms", "")) / 1000
                elif attr["time"].endswith("s"):
                    request.silence_seconds = float(attr["time"].replace("s", ""))
                else:
                    request.silence_seconds = 1
            elif attr.__contains__("strength"):
                if attr["strength"] == "small":
                    request.silence_seconds = .25
                elif attr["strength"] == "large":
                    request.silence_seconds = 2.5
                else:
                    request.silence_seconds = 1
            else:
                request.silence_seconds = 1

        else:
            for tag in prosody_tags:
                if tag.startswith("<prosody"):
                    attr = self.__extract_attributes(tag)
                    if attr.__contains__("volume"):
                        self.__apply_gain(request, attr["volume"])
                    if attr.__contains__("rate"):
                        self.__apply_rate(request, attr["rate"])
                    if attr.__contains__("pitch"):
                        self.__apply_pitch(request, attr["pitch"])
                elif tag.startswith("<voice"):
                    request.use_alt_voice = True

            request.message = message
        return request

    @staticmethod
    def __apply_gain(request: PendingSpeechRequest, volume: str):
        if volume.isnumeric():
            request.speech_settings.gain = float(volume) / 100.0 * 16 + -16
        elif volume == "silent":
            request.speech_settings.gain = -50
        elif volume == "x-soft":
            request.speech_settings.gain = -16
        elif volume == "soft":
            request.speech_settings.gain = -12
        elif volume == "medium":
            request.speech_settings.gain = -8
        elif volume == "loud":
            request.speech_settings.gain = -4
        else:
            request.speech_settings.gain = 0

    @staticmethod
    def __apply_rate(request: PendingSpeechRequest, rate: str):
        if rate.isnumeric():
            rate = float(rate)
            if rate < .5:
                request.speech_settings.speed = 0.5
            elif rate > 2:
                request.speech_settings.speed = 2
            else:
                request.speech_settings.speed = rate
        elif rate == "x-slow":
            request.speech_settings.speed = .5
        elif rate == "slow":
            request.speech_settings.speed = .8
        elif rate == "fast":
            request.speech_settings.speed = 1.55
        elif rate == "x-fast":
            request.speech_settings.speed *= 2

    @staticmethod
    def __apply_pitch(request: PendingSpeechRequest, pitch: str):
        if pitch.__contains__("%"):
            pitch = float(pitch.replace("%", "")) / 100.0
            if pitch < 0.5:
                request.speech_settings.pitch = 0.5
                request.speech_settings.speed *= 1.5
            elif pitch > 1.5:
                request.speech_settings.pitch = 1.5
                request.speech_settings.speed *= .5
            else:
                request.speech_settings.pitch = pitch
                request.speech_settings.speed *= 2 - pitch
        elif pitch == "x-low":
            request.speech_settings.pitch = 0.9
            request.speech_settings.speed *= 1.1
        elif pitch == "low":
            request.speech_settings.pitch = 0.95
            request.speech_settings.speed *= 1.05
        elif pitch == "high":
            request.speech_settings.pitch = 1.05
            request.speech_settings.speed *= 0.95
        elif pitch == "x-high":
            request.speech_settings.pitch = 1.1
            request.speech_settings.speed *= .9

    async def __send_response(self, request: PendingSpeechRequest, is_start: bool):
        if self.grpc_response_queue:
            response = speech_service_pb2.SpeechServiceResponse()
            response.speak_update.message = request.original_message
            response.speak_update.chunk = request.message
            response.speak_update.is_start_of_message = is_start and request.first_request_of_message
            response.speak_update.is_start_of_chunk = is_start
            response.speak_update.is_end_of_message = not is_start and (request.last_request_of_message or self.stop_talking_event.is_set())
            response.speak_update.is_end_of_chunk = not is_start
            await self.grpc_response_queue.put(response)

    async def __handle_request(self, request: PendingSpeechRequest):
        self.is_speaking = True

        if hasattr(request, "message") and request.message:

            await self.__send_response(request, True)

            speech_settings = self.default_settings
            if hasattr(request, "speech_settings"):
                speech_settings = request.speech_settings

            if hasattr(request, "use_alt_voice") and request.use_alt_voice:
                self.piper.set_speech_settings(speech_settings.alt_onnx_path, speech_settings.alt_config_path,
                                               speech_settings.alt_model_name)
            else:
                self.piper.set_speech_settings(speech_settings.onnx_path, speech_settings.config_path,
                                               speech_settings.model_name)

            sound = self.piper.text_to_pydub(request.message, speech_settings.speed)
            frame_rate_modifier = 1

            if speech_settings.gain != 0:
                sound = sound.apply_gain(speech_settings.gain)
            if speech_settings.pitch != 1:
                frame_rate_modifier = speech_settings.pitch

            new_frame_rate = int(sound.frame_rate * frame_rate_modifier)
            sound = sound._spawn(sound.raw_data, overrides={"frame_rate": new_frame_rate})

            p = pyaudio.PyAudio()
            stream = p.open(format=p.get_format_from_width(sound.sample_width),
                            channels=sound.channels,
                            rate=sound.frame_rate,
                            output=True)

            logging.info("Saying: " + str(request.message))

            try:
                for chunk in make_chunks(sound, 500):
                    if not self.stop_talking_event.is_set() and not self.shutdown_event.is_set():
                        await asyncio.to_thread(stream.write, chunk._data)

            finally:
                await self.__send_response(request, False)
                stream.stop_stream()
                stream.close()
                p.terminate()
                self.stop_talking_event.clear()
                self.is_speaking = False
        elif hasattr(request, "silence_seconds") and request.silence_seconds:
            await asyncio.sleep(request.silence_seconds)
            self.is_speaking = False


import asyncio
import datetime
import logging
import sys
import time
from asyncio import Server, Queue
from cgi import print_exception

from grpc import aio

from py_speech_service import speech_service_pb2_grpc, speech_service_pb2
from py_speech_service.speaker import Speaker, SpeechSettings
from py_speech_service.speech_recognition import SpeechRecognition


class GrpcServer:

    speaker: Speaker
    speech_recognition: SpeechRecognition
    server: Server
    shutdown_event = asyncio.Event()
    response_queue = Queue()

    def __init__(self):
        self.speaker = Speaker()
        self.speech_recognition = SpeechRecognition()

    async def start(self):
        server = aio.server()
        self.server = server
        speech_service_pb2_grpc.add_SpeechServiceServicer_to_server(self, server)
        port = server.add_insecure_port("[::]:0")
        await server.start()
        logging.info("Listening to gRPC connections on port " + str(port))
        print("grpc_port=" + str(port))
        sys.stdout.flush()
        self.speaker.start()
        await self.shutdown_event.wait()
        self.speaker.shutdown()
        if self.speech_recognition:
            self.speech_recognition.shutdown()
        await server.stop(5)
        time.sleep(1)

    async def StartSpeechService(self, request_iterator, context):
        self.speaker.set_grpc_response_queue(self.response_queue)
        self.speech_recognition.set_grpc_response_queue(self.response_queue)
        asyncio.create_task(self.process_queue(context))

        async for request in request_iterator:
            try:
                if request.HasField("start_speech_recognition"):
                    logging.info(
                        "Received gRPC start_speech_recognition request: " + str(request.start_speech_recognition))
                    print("Received gRPC start_speech_recognition request")

                    vosk_model = request.start_speech_recognition.vosk_model if hasattr(request.start_speech_recognition,
                        "vosk_model") else None
                    grammar_file = request.start_speech_recognition.grammar_file if hasattr(request.start_speech_recognition,
                        "grammar_file") else None
                    required_confidence = request.start_speech_recognition.required_confidence if hasattr(request.start_speech_recognition,
                        "required_confidence") else 80

                    self.speech_recognition.set_vosk_details(vosk_model, grammar_file, required_confidence)
                    asyncio.create_task(self.speech_recognition.start_speech_recognition(context))

                elif request.HasField("set_speech_defaults"):
                    logging.info(
                        "Received gRPC set_speech_defaults request: " + str(request.set_speech_defaults))
                    print("Received gRPC set_speech_defaults request")

                    speech_settings = SpeechSettings(request.set_speech_defaults.speech_settings)
                    alt_speech_settings = SpeechSettings(request.set_speech_defaults.alt_voice_settings) if request.set_speech_defaults.HasField(
                        "alt_voice_settings") else request.set_speech_defaults.speech_settings
                    self.speaker.default_settings = SpeechSettings(speech_settings)
                    self.speaker.alt_voice_settings = alt_speech_settings
                elif request.HasField("speak"):
                    logging.info(
                        "Received gRPC speak request: " + str(request.speak))
                    print("Received gRPC speak request")

                    speech_settings = SpeechSettings(request.speak.speech_settings) if request.speak.HasField("speech_settings") else None
                    await self.speaker.speak(request.speak.message, speech_settings)
                elif request.HasField("stop_speaking"):
                    logging.info(
                        "Received gRPC stop_speaking request")
                    print("Received gRPC stop_speaking request")

                    self.speaker.stop_speaking()
                elif request.HasField("shutdown"):
                    logging.info(
                        "Received gRPC shutdown request")
                    print("Received gRPC shutdown request")
                    break
                elif request.HasField("ping"):
                    print("Received gRPC ping")
                    logging.info(
                        "Received gRPC ping")
                    response = speech_service_pb2.SpeechServiceResponse()
                    #response.ping = speech_service_pb2.PingResponse()
                    response.ping.time = str(datetime.datetime.now())
                    await context.write(response)
            except Exception as e:
                logging.error(e)
                print_exception(value = e)
                error_details = speech_service_pb2.SpeechServiceError()
                error_details.original_request = request
                error_details.message = repr(e)
                response = speech_service_pb2.SpeechServiceResponse()
                response.error = error_details
                await self.response_queue.put(response)
                logging.error(e)

        self.shutdown_event.set()

    async def process_queue(self, context):
        while not self.shutdown_event.is_set():
            response = await self.response_queue.get()
            await context.write(response)





import json, re
import logging
import typing

import num2words
from rapidfuzz import process
from py_speech_service.grammar_element import GrammarRuleElement, GrammarElementType, GrammarElementLookupItem, \
    GrammarElementMatch

class GrammarParser:

    phrase_map: dict[str, list[GrammarElementLookupItem]] = {}
    leading_phrases: list[str] = []
    pattern = re.compile('[^\\w ]+')
    max_phrase_word_count: int = 0
    all_words: list[str] = []

    def set_grammar_file(self, file_path: str):
        with open(file_path, 'r') as fp:
            lines = fp.read()

        json_data = json.loads(lines)
        for rule in json_data:
            rule_name: str = rule["Key"]
            self.__parse_rule_element(rule_name, rule)

        self.all_words = list(set(self.all_words))
        updated_words = []
        for word in self.all_words:
            updated_words += self.pattern.sub('', word).lower().split()
        self.all_words = list(set(updated_words))

    def find_match(self, stated_text: str, min_threshold: float = 80, min_prefix_threshold: float = 60):
        search_text = self.pattern.sub('', stated_text).lower().strip()
        search_words = search_text.split()
        search_word_count = len(search_words)
        if search_word_count > 30:
            return None

        prefix_search = " ".join(search_words[:2])
        prefix_result = self.__find_closest_sentence(["hey tracker"], prefix_search)
        if prefix_result is None:
            return None
        if prefix_result[1] < min_prefix_threshold:
            return None
        search_words[0] = "hey"
        search_words[1] = "tracker"
        search_text = " ".join(search_words)

        highest_confidence: float = 0
        highest_match: typing.Optional[GrammarElementLookupItem] = None
        possibilities: [(str, float)] = []
        for i in range(search_word_count, 1, -1):
            search_phrase = " ".join(search_words[0:i])
            search_result = self.__find_closest_sentence(self.leading_phrases, search_phrase)
            if search_result is not None and search_result[1] > min_threshold:
                possibilities.append(search_result)

        matches: [GrammarElementMatch] = []
        searched_phrases: [str] = []
        for possibility in possibilities:
            search_phrase = possibility[0]
            if searched_phrases.__contains__(search_phrase):
                continue
            searched_phrases.append(search_phrase)
            initial_confidence = possibility[1]
            possible_elements = self.phrase_map.get(search_phrase)
            searched_elements = []
            for possible_element in possible_elements:
                if searched_elements.__contains__(possible_element.rule_name + possible_element.phrase):
                    continue
                searched_elements.append(possible_element.rule_name + possible_element.phrase)
                if possible_element.is_full_match:
                    if initial_confidence > min_threshold:
                        match = GrammarElementMatch(possible_element.rule_name, stated_text, search_phrase, initial_confidence)
                        matches.append(match)
                else:
                    best_item = self.__find_best_element(search_text, search_phrase, possible_element)
                    if best_item is not None and best_item[1] > min_threshold:
                        match = GrammarElementMatch(possible_element.rule_name, stated_text, best_item[0],
                                                    best_item[1], best_item[2])
                        matches.append(match)

        if len(matches) == 0:
            return None
        else:
            best_match: typing.Optional[GrammarElementMatch] = None
            longest_match: typing.Optional[GrammarElementMatch] = None
            for match in matches:
                if best_match is None:
                    best_match = match
                    longest_match = match
                else:
                    if len(match.matched_text) > len(best_match.matched_text):
                        longest_match = match
                    if match.confidence > best_match.confidence:
                        best_match = match

            if longest_match.confidence >= best_match.confidence - 5:
                return longest_match
            else:
                return best_match

    def __parse_rule_element(self, rule_name: str, rule_element):

        element = GrammarRuleElement()
        element.rule = rule_name
        element.data = []
        element_phrases: list[str] = [""]

        is_exact: bool = True
        words = []

        for sub_element_json in rule_element['Data']:
            element.data.append(sub_element_json)

            element_type: GrammarElementType = GrammarElementType(sub_element_json['Type'])

            if element_type == GrammarElementType.String:
                for i in range(0, len(element_phrases)):
                    text = self.pattern.sub('', sub_element_json['Data'].strip() + " ").lower()
                    words += [ text]
                    if is_exact:
                        element_phrases[i] += text
            elif element_type == GrammarElementType.OneOf:
                one_of_phrases: [str] = sub_element_json['Data']
                words += one_of_phrases
                if is_exact:
                    element_phrases = self.__permutate_elements(element_phrases, one_of_phrases)
            elif element_type == GrammarElementType.Optional:
                optional_phrases: [str] = sub_element_json['Data']
                words += optional_phrases
                if is_exact:
                    optional_phrases.append("")
                    element_phrases = self.__permutate_elements(element_phrases, optional_phrases)
            elif element_type == GrammarElementType.KeyValue:
                for key_value_json in sub_element_json['Data']:
                    key = key_value_json['Key']
                    if key.isnumeric():
                        key = num2words.num2words(key).replace("-", " ")
                    words.append(key)
                is_exact = False
            elif element_type == GrammarElementType.GrammarElementList:
                for grammar_list_item in sub_element_json['Data']:
                    self.__parse_rule_element(rule_name, grammar_list_item)
                return

        self.all_words += words
        for phrase in element_phrases:
            match_details = GrammarElementLookupItem(rule_name, phrase, element, is_exact)
            word_count = len(phrase.split())
            if phrase.strip() == "" or word_count < 2:
                continue
            if word_count > self.max_phrase_word_count:
                self.max_phrase_word_count = word_count
            if self.phrase_map.__contains__(phrase):
                self.phrase_map[phrase].append(match_details)
            else:
                self.phrase_map[phrase] = [ match_details ]
                self.leading_phrases.append(phrase)

    def __permutate_elements(self, initial_items: [str], additional_items: [str]):
        initial_count = len(initial_items)
        to_return = self.__duplicate_elements(initial_items, len(additional_items))
        index = 0
        for i in range(0, initial_count):
            for text in additional_items:
                text = self.pattern.sub('', text.strip() + " ").lower()
                if text == " ":
                    index = index + 1
                    continue
                to_return[index] += text
                index = index + 1
        return to_return

    def __find_best_element(self, search_text: str, current_phrase: str, element: GrammarElementLookupItem):
        hit_key_value = False
        search_phrase = current_phrase
        selected_values: dict[str, str] = {}
        confidence = 0
        search_words = search_text.split()

        for sub_element_json in element.grammar_element.data:
            element_type: GrammarElementType = GrammarElementType(sub_element_json['Type'])
            if not hit_key_value:
                if element_type == GrammarElementType.KeyValue:
                    hit_key_value = True
                else:
                    continue

            if element_type == GrammarElementType.String:
                text = self.pattern.sub('', sub_element_json['Data'].strip() + " ").lower()
                search_phrase += text
            elif element_type == GrammarElementType.OneOf:
                items: [str] = sub_element_json['Data']
                match = self.__find_best_element_sub_phrase(search_phrase, search_words, items)
                if match is None:
                    return None
                else:
                    search_phrase = match[0]
                    confidence = match[1]
            elif element_type == GrammarElementType.Optional:
                items: [str] = sub_element_json['Data']
                match = self.__find_best_element_sub_phrase(search_phrase, search_words, items)
                if match is not None and match[1] > confidence:
                    search_phrase = match[0]
                    confidence = match[1]
            elif element_type == GrammarElementType.KeyValue:
                keys: [str] = []
                key_values: dict[str, str] = {}
                for key_value_json in sub_element_json['Data']:
                    key = key_value_json['Key']
                    if key.isnumeric():
                        key = num2words.num2words(key).replace("-", " ")
                    keys.append(key)
                    key_values[key] = key_value_json['Value']
                match = self.__find_best_element_sub_phrase(search_phrase, search_words, keys)
                if match is None:
                    return None
                else:
                    search_phrase = match[0]
                    confidence = match[1]
                    selected_values[str(sub_element_json['Key'])] = str(key_values[match[2]])

        return search_phrase, confidence, selected_values

    def __find_best_element_sub_phrase(self, search_phrase: str, search_words: [str], items: [str]):
        if not search_phrase.endswith(" "):
            search_phrase = search_phrase + " "
        possible_phrases: [str] = []
        phrase_addition_map: dict[str, str] = {}
        max_word_count = 0
        min_word_count = 10000
        for item in items:
            text = self.pattern.sub('', item.strip() + " ").lower()
            new_phrase = search_phrase + text
            possible_phrases.append(new_phrase)
            phrase_addition_map[new_phrase] = item
            new_phrase_word_count = len(new_phrase.split())
            if new_phrase_word_count < min_word_count:
                min_word_count = new_phrase_word_count
            if new_phrase_word_count > max_word_count:
                max_word_count = new_phrase_word_count

        best_result = None
        previous_search_phrase = ""
        for num_words in range(min_word_count, max_word_count+1):
            new_search_phrase = " ".join(search_words[0:num_words])
            if new_search_phrase == previous_search_phrase:
                break
            previous_search_phrase = new_search_phrase
            result = self.__find_closest_sentence(possible_phrases, new_search_phrase)
            if result is not None:
                if best_result is None:
                    best_result = result
                elif result[1] >= best_result[1] - 5:
                    best_result = result

        if best_result is None:
            return None

        return best_result[0], best_result[1], phrase_addition_map[best_result[0]]

    def __find_closest_sentence(self, sentences: [str], query: str) -> (str, float):
        filtered_sentences = self.__filter_by_length(sentences, query)
        squashed_sentences = []
        squashed_map = {}
        for sentence in filtered_sentences:
            squashed_sentence = sentence.replace(" ", "")
            squashed_sentences.append(squashed_sentence)
            squashed_map[squashed_sentence] = sentence
        response = process.extractOne(query.replace(" ", ""), squashed_sentences)
        if response is None:
            return None
        return squashed_map[response[0]], response[1]

    @staticmethod
    def __filter_by_length(items: [], reference_string: str, tolerance: int = 4):
        reference_length = len(reference_string)
        return [s for s in items if abs(len(s) - reference_length) <= tolerance]

    @staticmethod
    def __duplicate_elements(items: [], times: int):
        if times < 1:
            return []  # Return an empty list if times is less than 1
        return [element for element in items for _ in range(times)]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: speech_service.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'speech_service.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14speech_service.proto\"\xb7\x02\n\x14SpeechServiceRequest\x12\x42\n\x18start_speech_recognition\x18\x01 \x01(\x0b\x32\x1e.StartSpeechRecognitionRequestH\x00\x12\x38\n\x13set_speech_defaults\x18\x02 \x01(\x0b\x32\x19.SetSpeechDefaultsRequestH\x00\x12\x1e\n\x05speak\x18\x03 \x01(\x0b\x32\r.SpeakRequestH\x00\x12-\n\rstop_speaking\x18\x04 \x01(\x0b\x32\x14.StopSpeakingRequestH\x00\x12$\n\x08shutdown\x18\x05 \x01(\x0b\x32\x10.ShutdownRequestH\x00\x12\x1c\n\x04ping\x18\x06 \x01(\x0b\x32\x0c.PingRequestH\x00\x42\x0e\n\x0cmessage_type\"\xd3\x01\n\x15SpeechServiceResponse\x12,\n\x0cspeak_update\x18\x01 \x01(\x0b\x32\x14.SpeakUpdateResponseH\x00\x12\x37\n\x11speech_recognized\x18\x02 \x01(\x0b\x32\x1a.SpeechRecognitionResponseH\x00\x12$\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x13.SpeechServiceErrorH\x00\x12\x1d\n\x04ping\x18\x04 \x01(\x0b\x32\r.PingResponseH\x00\x42\x0e\n\x0cmessage_type\"\\\n\x12SpeechServiceError\x12/\n\x10original_request\x18\x01 \x01(\x0b\x32\x15.SpeechServiceRequest\x12\x15\n\rerror_message\x18\x02 \x01(\t\"f\n\x1dStartSpeechRecognitionRequest\x12\x12\n\nvosk_model\x18\x01 \x01(\t\x12\x14\n\x0cgrammar_file\x18\x02 \x01(\t\x12\x1b\n\x13required_confidence\x18\x03 \x01(\x01\"\x8d\x01\n\x18SetSpeechDefaultsRequest\x12(\n\x0fspeech_settings\x18\x01 \x01(\x0b\x32\x0f.SpeechSettings\x12\x30\n\x12\x61lt_voice_settings\x18\x02 \x01(\x0b\x32\x0f.SpeechSettingsH\x00\x88\x01\x01\x42\x15\n\x13_alt_voice_settings\"\xc0\x01\n\x0eSpeechSettings\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x11\n\tonnx_path\x18\x02 \x01(\t\x12\x13\n\x0b\x63onfig_path\x18\x03 \x01(\t\x12\x16\n\x0e\x61lt_model_name\x18\x04 \x01(\t\x12\x15\n\ralt_onnx_path\x18\x05 \x01(\t\x12\x17\n\x0f\x61lt_config_path\x18\x06 \x01(\t\x12\r\n\x05speed\x18\x07 \x01(\x01\x12\x0c\n\x04gain\x18\x08 \x01(\x01\x12\r\n\x05pitch\x18\t \x01(\x01\"b\n\x0cSpeakRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\x12-\n\x0fspeech_settings\x18\x02 \x01(\x0b\x32\x0f.SpeechSettingsH\x00\x88\x01\x01\x42\x12\n\x10_speech_settings\"\x15\n\x13StopSpeakingRequest\"\xa1\x01\n\x13SpeakUpdateResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\r\n\x05\x63hunk\x18\x02 \x01(\t\x12\x1b\n\x13is_start_of_message\x18\x03 \x01(\x08\x12\x19\n\x11is_start_of_chunk\x18\x04 \x01(\x08\x12\x19\n\x11is_end_of_message\x18\x05 \x01(\x08\x12\x17\n\x0fis_end_of_chunk\x18\x06 \x01(\x08\"\xe5\x01\n\x19SpeechRecognitionResponse\x12\x12\n\nheard_text\x18\x01 \x01(\t\x12\x17\n\x0frecognized_text\x18\x02 \x01(\t\x12\x17\n\x0frecognized_rule\x18\x03 \x01(\t\x12\x12\n\nconfidence\x18\x04 \x01(\x01\x12<\n\tsemantics\x18\x05 \x03(\x0b\x32).SpeechRecognitionResponse.SemanticsEntry\x1a\x30\n\x0eSemanticsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x11\n\x0fShutdownRequest\"\x1b\n\x0bPingRequest\x12\x0c\n\x04time\x18\x01 \x01(\t\"\x1c\n\x0cPingResponse\x12\x0c\n\x04time\x18\x01 \x01(\t2X\n\rSpeechService\x12G\n\x12StartSpeechService\x12\x15.SpeechServiceRequest\x1a\x16.SpeechServiceResponse(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'speech_service_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY']._loaded_options = None
  _globals['_SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY']._serialized_options = b'8\001'
  _globals['_SPEECHSERVICEREQUEST']._serialized_start=25
  _globals['_SPEECHSERVICEREQUEST']._serialized_end=336
  _globals['_SPEECHSERVICERESPONSE']._serialized_start=339
  _globals['_SPEECHSERVICERESPONSE']._serialized_end=550
  _globals['_SPEECHSERVICEERROR']._serialized_start=552
  _globals['_SPEECHSERVICEERROR']._serialized_end=644
  _globals['_STARTSPEECHRECOGNITIONREQUEST']._serialized_start=646
  _globals['_STARTSPEECHRECOGNITIONREQUEST']._serialized_end=748
  _globals['_SETSPEECHDEFAULTSREQUEST']._serialized_start=751
  _globals['_SETSPEECHDEFAULTSREQUEST']._serialized_end=892
  _globals['_SPEECHSETTINGS']._serialized_start=895
  _globals['_SPEECHSETTINGS']._serialized_end=1087
  _globals['_SPEAKREQUEST']._serialized_start=1089
  _globals['_SPEAKREQUEST']._serialized_end=1187
  _globals['_STOPSPEAKINGREQUEST']._serialized_start=1189
  _globals['_STOPSPEAKINGREQUEST']._serialized_end=1210
  _globals['_SPEAKUPDATERESPONSE']._serialized_start=1213
  _globals['_SPEAKUPDATERESPONSE']._serialized_end=1374
  _globals['_SPEECHRECOGNITIONRESPONSE']._serialized_start=1377
  _globals['_SPEECHRECOGNITIONRESPONSE']._serialized_end=1606
  _globals['_SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY']._serialized_start=1558
  _globals['_SPEECHRECOGNITIONRESPONSE_SEMANTICSENTRY']._serialized_end=1606
  _globals['_SHUTDOWNREQUEST']._serialized_start=1608
  _globals['_SHUTDOWNREQUEST']._serialized_end=1625
  _globals['_PINGREQUEST']._serialized_start=1627
  _globals['_PINGREQUEST']._serialized_end=1654
  _globals['_PINGRESPONSE']._serialized_start=1656
  _globals['_PINGRESPONSE']._serialized_end=1684
  _globals['_SPEECHSERVICE']._serialized_start=1686
  _globals['_SPEECHSERVICE']._serialized_end=1774
# @@protoc_insertion_point(module_scope)



import asyncio
import json
import logging
from typing import Optional

from vosk import Model, KaldiRecognizer, SetLogLevel

from py_speech_service import speech_service_pb2
from py_speech_service.grammar_parser import GrammarParser

import queue
import sounddevice as sd

class SpeechRecognition:
    model: Model
    grammar_parser: GrammarParser
    required_confidence: float = 80
    stop_speech_recognition_event: Optional[asyncio.Event] = None
    continue_speech_recognition: bool = True
    grpc_response_queue: Optional[asyncio.Queue] = None
    shutdown_event = asyncio.Event()
    recognition_queue = asyncio.Queue()

    def __init__(self):
        SetLogLevel(-1)
        self.stop_speech_recognition_event = None

    def set_vosk_details(self, vosk_model: str, grammar_file: str, required_confidence: float = 80):
        self.grammar_parser = GrammarParser()
        self.grammar_parser.set_grammar_file(grammar_file)

        if vosk_model.count("/") > 0 or vosk_model.count("\\") > 0:
            self.model = Model(vosk_model)
        else:
            self.model = Model(None, vosk_model)

        self.required_confidence = required_confidence

    async def start_speech_recognition(self, context):
        asyncio.create_task(self.process_recognition_queue())
        await asyncio.to_thread(self.listen)

    async def process_recognition_queue(self):
        logging.info("Started processing speech recognition queue")
        print("Started processing speech recognition queue")
        while not self.shutdown_event.is_set():  # Keep running unless shutdown is triggered
            try:
                item = await asyncio.wait_for(self.recognition_queue.get(), timeout=1.0)  # Wait for 1 second
                await self.process_speech(item)
            except asyncio.TimeoutError:
                continue  # Retry checking
        self.recognition_queue.task_done()
        logging.info("Stopped processing speech recognition queue")
        print("Stopped processing speech recognition queue")

    def listen(self):
        if self.stop_speech_recognition_event:
            self.stop_speech_recognition_event.set()

        stop_speech_recognition_event = asyncio.Event()
        self.stop_speech_recognition_event = stop_speech_recognition_event

        device_info = sd.query_devices(sd.default.device[0], 'input')
        samplerate = int(device_info['default_samplerate'])
        speech_queue = queue.Queue()

        def record_callback(indata, frames, time, status):
            if status:
                logging.error(str(status))
            speech_queue.put(bytes(indata))

        words_json = json.dumps(self.grammar_parser.all_words)
        recognizer = KaldiRecognizer(self.model, samplerate, words_json)
        recognizer.SetWords(False)
        recognizer.SetGrammar(words_json)

        try:
            logging.info("Started listening to voice via VOSK")
            print("Started listening to voice via VOSK")
            with sd.RawInputStream(dtype='int16',
                                   channels=1,
                                   callback=record_callback):
                while not stop_speech_recognition_event.is_set() and not self.shutdown_event.is_set():
                    data = speech_queue.get()
                    if recognizer.AcceptWaveform(data):
                        recognized_text = recognizer.Result()
                        if recognized_text:
                            asyncio.run(self.recognition_queue.put(recognized_text))
            logging.info("Stopped listening to voice via VOSK")
            print("Stopped listening to voice via VOSK")
        except KeyboardInterrupt:
            print('Finished recording due to keyboard interrupt')
            logging.error("Finished recording due to keyboard interrupt")
        except Exception as e:
            print("Error from VOSK: " + str(e))
            logging.error("Error from VOSK")
            logging.error(e)

    async def process_speech(self, recognizer_result: str):
        try:
            result_dict = json.loads(recognizer_result)
            text_recognized = result_dict.get("text", "")
            if not text_recognized == "":
                match = self.grammar_parser.find_match(text_recognized, self.required_confidence)

                if match is not None:
                    if self.grpc_response_queue:
                        response = speech_service_pb2.SpeechServiceResponse()
                        response.speech_recognized.heard_text = text_recognized
                        response.speech_recognized.recognized_text = match.matched_text
                        response.speech_recognized.recognized_rule = match.rule
                        response.speech_recognized.confidence = round(match.confidence, 2)
                        response.speech_recognized.semantics.update(match.values)
                        await self.grpc_response_queue.put(response)
                    else:
                        print("I heard: '" + text_recognized + "', but I am " + str(round(match.confidence, 2)) + "% sure you said '" + match.matched_text + "'")

        except Exception as e:
            print("Error processing speech: " + str(e))
            logging.error("Error processing speech")
            logging.error(e)

    def set_grpc_response_queue(self, queue: asyncio.Queue):
        self.grpc_response_queue = queue

    def stop_speech_recognition(self):
        if self.stop_speech_recognition_event is not None:
            self.stop_speech_recognition_event.set()

    def shutdown(self):
        self.shutdown_event.set()
